@@ -20,7 +20,9 @@ from eventlet import tpool
 from eventlet import wsgi
 from eventlet.green import socket as greensocket
 from eventlet.green import ssl
-from eventlet.support import bytes_to_str, six
+from eventlet.support import bytes_to_str
+import six
+from six.moves.urllib import parse
 import tests
 
 
@@ -101,7 +103,8 @@ def chunked_post(env, start_response):
 
 def already_handled(env, start_response):
     start_response('200 OK', [('Content-type', 'text/plain')])
-    return wsgi.ALREADY_HANDLED
+    wsgi.WSGI_LOCAL.already_handled = True
+    return []
 
 
 class Site(object):
@@ -113,8 +116,7 @@ class Site(object):
 
 
 class IterableApp(object):
-
-    def __init__(self, send_start_response=False, return_val=wsgi.ALREADY_HANDLED):
+    def __init__(self, send_start_response=False, return_val=()):
         self.send_start_response = send_start_response
         self.return_val = return_val
         self.env = {}
@@ -123,6 +125,8 @@ class IterableApp(object):
         self.env = env
         if self.send_start_response:
             start_response('200 OK', [('Content-type', 'text/plain')])
+        else:
+            wsgi.WSGI_LOCAL.already_handled = True
         return self.return_val
 
 
@@ -188,7 +192,7 @@ def read_http(sock):
         x = x.strip()
         if not x:
             continue
-        key, value = bytes_to_str(x).split(':', 1)
+        key, value = bytes_to_str(x, encoding='latin1').split(':', 1)
         key = key.rstrip()
         value = value.lstrip()
         key_lower = key.lower()
@@ -317,7 +321,7 @@ class TestHttpd(_TestBase):
         # define a new handler that does a get_arg as well as a read_body
         def new_app(env, start_response):
             body = bytes_to_str(env['wsgi.input'].read())
-            a = cgi.parse_qs(body).get('a', [1])[0]
+            a = parse.parse_qs(body).get('a', [1])[0]
             start_response('200 OK', [('Content-type', 'text/plain')])
             return [six.b('a is %s, body is %s' % (a, body))]
 
@@ -729,6 +733,13 @@ class TestHttpd(_TestBase):
         assert b'400 Bad Request' in result, result
         assert b'500' not in result, result
 
+        sock = eventlet.connect(self.server_addr)
+        sock.sendall(b'GET / HTTP/1.0\r\nHost: localhost\r\nContent-length: -10\r\n\r\n')
+        result = recvall(sock)
+        assert result.startswith(b'HTTP'), result
+        assert b'400 Bad Request' in result, result
+        assert b'500' not in result, result
+
     def test_024_expect_100_continue(self):
         def wsgi_app(environ, start_response):
             if int(environ['CONTENT_LENGTH']) > 1024:
@@ -747,27 +758,29 @@ class TestHttpd(_TestBase):
         result = read_http(sock)
         self.assertEqual(result.status, 'HTTP/1.1 417 Expectation Failed')
         self.assertEqual(result.body, b'failure')
-        fd.write(
-            b'PUT / HTTP/1.1\r\nHost: localhost\r\nContent-length: 7\r\n'
-            b'Expect: 100-continue\r\n\r\ntesting')
-        fd.flush()
-        header_lines = []
-        while True:
-            line = fd.readline()
-            if line == b'\r\n':
-                break
-            else:
-                header_lines.append(line)
-        assert header_lines[0].startswith(b'HTTP/1.1 100 Continue')
-        header_lines = []
-        while True:
-            line = fd.readline()
-            if line == b'\r\n':
-                break
-            else:
-                header_lines.append(line)
-        assert header_lines[0].startswith(b'HTTP/1.1 200 OK')
-        assert fd.read(7) == b'testing'
+
+        for expect_value in ('100-continue', '100-Continue'):
+            fd.write(
+                'PUT / HTTP/1.1\r\nHost: localhost\r\nContent-length: 7\r\n'
+                'Expect: {}\r\n\r\ntesting'.format(expect_value).encode())
+            fd.flush()
+            header_lines = []
+            while True:
+                line = fd.readline()
+                if line == b'\r\n':
+                    break
+                else:
+                    header_lines.append(line)
+            assert header_lines[0].startswith(b'HTTP/1.1 100 Continue')
+            header_lines = []
+            while True:
+                line = fd.readline()
+                if line == b'\r\n':
+                    break
+                else:
+                    header_lines.append(line)
+            assert header_lines[0].startswith(b'HTTP/1.1 200 OK')
+            assert fd.read(7) == b'testing'
         fd.close()
         sock.close()
 
@@ -978,6 +991,67 @@ class TestHttpd(_TestBase):
         fd.close()
         sock.close()
 
+    def test_024d_expect_100_continue_with_eager_app_chunked(self):
+        def wsgi_app(environ, start_response):
+            # app knows it's going to do some time-intensive thing and doesn't
+            # want clients to time out, so it's protocol says to:
+            # * generally expect a successful status code,
+            # * be prepared to eat some whitespace that will get dribbled out
+            #   periodically, and
+            # * parse the final status from the response body.
+            environ['eventlet.minimum_write_chunk_size'] = 0
+            start_response('202 Accepted', [])
+
+            def resp_gen():
+                yield b' '
+                environ['wsgi.input'].read()
+                yield b' '
+                yield b'\n503 Service Unavailable\n\nOops!\n'
+
+            return resp_gen()
+
+        self.site.application = wsgi_app
+        sock = eventlet.connect(self.server_addr)
+        fd = sock.makefile('rwb')
+        fd.write(b'PUT /a HTTP/1.1\r\n'
+                 b'Host: localhost\r\nConnection: close\r\n'
+                 b'Transfer-Encoding: chunked\r\n'
+                 b'Expect: 100-continue\r\n\r\n')
+        fd.flush()
+
+        # Expect the optimistic response
+        header_lines = []
+        while True:
+            line = fd.readline()
+            if line == b'\r\n':
+                break
+            else:
+                header_lines.append(line.strip())
+        self.assertEqual(header_lines[0], b'HTTP/1.1 202 Accepted')
+
+        def chunkify(data):
+            return '{:x}'.format(len(data)).encode('ascii') + b'\r\n' + data + b'\r\n'
+
+        def expect_chunk(data):
+            expected = chunkify(data)
+            self.assertEqual(expected, fd.read(len(expected)))
+
+        # Can even see that initial whitespace
+        expect_chunk(b' ')
+
+        # Send message
+        fd.write(chunkify(b'some data'))
+        fd.write(chunkify(b''))  # end-of-message
+        fd.flush()
+
+        # Expect final response
+        expect_chunk(b' ')
+        expect_chunk(b'\n503 Service Unavailable\n\nOops!\n')
+        expect_chunk(b'')  # end-of-message
+
+        fd.close()
+        sock.close()
+
     def test_025_accept_errors(self):
         debug.hub_exceptions(True)
         listener = greensocket.socket()
@@ -1356,6 +1430,32 @@ class TestHttpd(_TestBase):
         self.assertEqual(read_content.wait(), b'ok')
         assert blew_up[0]
 
+    def test_aborted_post_io_error(self):
+        ran_post_req_hook = [False]
+
+        def post_req_hook(env):
+            ran_post_req_hook[0] = True
+
+        def early_responder(env, start_response):
+            env['eventlet.posthooks'] = [(post_req_hook, (), {})]
+            start_response('200 OK', [('Content-Type', 'text/plain')])
+            return ['ok']
+
+        self.site.application = early_responder
+        data = "\r\n".join(['PUT /somefile HTTP/1.1',
+                            'Transfer-Encoding: chunked',
+                            '',
+                            '20',
+                            'not 32 bytes'])
+        sock = eventlet.connect(self.server_addr)
+        sock.sendall(data.encode())
+        sock.close()
+        # Give the server a chance to wrap things up
+        eventlet.sleep(0.01)
+        # Unexpected EOF shouldn't kill the server;
+        # post-request processing should still happen
+        assert ran_post_req_hook[0]
+
     def test_exceptions_close_connection(self):
         def wsgi_app(environ, start_response):
             raise RuntimeError("intentional error")
@@ -1414,16 +1514,28 @@ class TestHttpd(_TestBase):
 
         self.site.application = wsgi_app
         sock = eventlet.connect(self.server_addr)
-        sock.sendall(b'GET /%E4%BD%A0%E5%A5%BD HTTP/1.1\r\nHost: localhost\r\nConnection: close\r\n\r\n')
+        # This is a properly-quoted request for the UTF-8 path /你好
+        sock.sendall(b'GET /%E4%BD%A0%E5%A5%BD HTTP/1.1\r\nHost: localhost\r\n\r\n')
         result = read_http(sock)
         assert result.status == 'HTTP/1.1 200 OK'
-        # that was only preparation, actual test below
+        # Like above, but the octets are reversed before being quoted,
+        # so the result should *not* be interpreted as UTF-8
+        sock.sendall(b'GET /%BD%A5%E5%A0%BD%E4 HTTP/1.1\r\nHost: localhost\r\nConnection: close\r\n\r\n')
+        result = read_http(sock)
+        assert result.status == 'HTTP/1.1 200 OK'
+
+        # that was only preparation, actual tests below
         # Per PEP-0333 https://www.python.org/dev/peps/pep-0333/#unicode-issues
         # in all WSGI environment strings application must observe either bytes in latin-1 (ISO-8859-1)
         # or unicode code points \u0000..\u00ff
-        # wsgi_decoding_dance from Werkzeug to emulate concerned application
-        decoded = g[0].encode('latin1').decode('utf-8', 'replace')
-        assert decoded == u'/你好'
+        msg = 'Expected PATH_INFO to be a native string, not {0}'.format(type(g[0]))
+        assert isinstance(g[0], str), msg
+        # Fortunately, WSGI strings have the same literal representation on both py2 and py3
+        assert g[0] == '/\xe4\xbd\xa0\xe5\xa5\xbd'
+
+        msg = 'Expected PATH_INFO to be a native string, not {0}'.format(type(g[1]))
+        assert isinstance(g[1], str), msg
+        assert g[1] == '/\xbd\xa5\xe5\xa0\xbd\xe4'
 
     @tests.skip_if_no_ipv6
     def test_ipv6(self):
@@ -1519,6 +1631,26 @@ class TestHttpd(_TestBase):
         except ConnectionClosed:
             pass
 
+    def test_header_name_capitalization(self):
+        def wsgi_app(environ, start_response):
+            start_response('200 oK', [
+                ('sOMe-WEirD', 'cAsE'),
+                ('wiTH-\xdf-LATIN1-\xff', 'chars'),
+            ])
+            return [b'']
+
+        self.spawn_server(site=wsgi_app)
+
+        sock = eventlet.connect(self.server_addr)
+        sock.sendall(b'GET / HTTP/1.1\r\nHost: localhost\r\n\r\n')
+        result = read_http(sock)
+        sock.close()
+        self.assertEqual(result.status, 'HTTP/1.1 200 oK')
+        self.assertIn('Some-Weird', result.headers_original)
+        self.assertEqual(result.headers_original['Some-Weird'], 'cAsE')
+        self.assertIn('With-\xdf-Latin1-\xff', result.headers_original)
+        self.assertEqual(result.headers_original['With-\xdf-Latin1-\xff'], 'chars')
+
     def test_disable_header_name_capitalization(self):
         # Disable HTTP header name capitalization
         #
@@ -1597,6 +1729,34 @@ class TestHttpd(_TestBase):
         assert result.body == (b'HTTP_HOST: localhost\nHTTP_HTTP_X_ANY_K: two\n'
                                b'HTTP_PATH_INFO: foo\nHTTP_X_ANY_K: one\n')
 
+    def test_env_header_stripping(self):
+        def app(environ, start_response):
+            start_response('200 OK', [])
+            # On py3, headers get parsed as Latin-1, so send them back out as Latin-1, too
+            return [line if isinstance(line, bytes) else line.encode('latin1')
+                    for kv in sorted(environ.items())
+                    if kv[0].startswith('HTTP_')
+                    for line in ('{0}: {1}\n'.format(*kv),)]
+
+        self.spawn_server(site=app)
+        sock = eventlet.connect(self.server_addr)
+        sock.sendall(
+            b'GET / HTTP/1.1\r\n'
+            b'Host: localhost\r\n'
+            b'spaced:   o   u   t   \r\n'
+            b'trailing: tab\t\r\n'
+            b'trailing-nbsp: \xc2\xa0\r\n'
+            b'null-set: \xe2\x88\x85\r\n\r\n')
+        result = read_http(sock)
+        sock.close()
+        assert result.status == 'HTTP/1.1 200 OK', 'Received status {0!r}'.format(result.status)
+        assert result.body == (
+            b'HTTP_HOST: localhost\n'
+            b'HTTP_NULL_SET: \xe2\x88\x85\n'
+            b'HTTP_SPACED: o   u   t\n'
+            b'HTTP_TRAILING: tab\n'
+            b'HTTP_TRAILING_NBSP: \xc2\xa0\n')
+
     def test_log_disable(self):
         self.spawn_server(log_output=False)
         sock = eventlet.connect(self.server_addr)
@@ -1704,6 +1864,12 @@ class TestChunkedInput(_TestBase):
         elif pi == "/lines":
             for x in input:
                 response.append(x)
+        elif pi == "/readline":
+            response.extend(iter(input.readline, b''))
+            response.append(('\nread %d lines' % len(response)).encode())
+        elif pi == "/readlines":
+            response.extend(input.readlines())
+            response.append(('\nread %d lines' % len(response)).encode())
         elif pi == "/ping":
             input.read()
             response.append(b"pong")
@@ -1807,6 +1973,26 @@ class TestChunkedInput(_TestBase):
         self.assertEqual(read_http(fd).body, b'this is chunked\nline 2\nline3')
         fd.close()
 
+    def test_chunked_readline_from_input(self):
+        body = self.body()
+        req = "POST /readline HTTP/1.1\r\nContent-Length: %s\r\n" \
+              "transfer-encoding: Chunked\r\n\r\n%s" % (len(body), body)
+
+        fd = self.connect()
+        fd.sendall(req.encode())
+        self.assertEqual(read_http(fd).body, b'this is chunked\nline 2\nline3\nread 3 lines')
+        fd.close()
+
+    def test_chunked_readlines_from_input(self):
+        body = self.body()
+        req = "POST /readlines HTTP/1.1\r\nContent-Length: %s\r\n" \
+              "transfer-encoding: Chunked\r\n\r\n%s" % (len(body), body)
+
+        fd = self.connect()
+        fd.sendall(req.encode())
+        self.assertEqual(read_http(fd).body, b'this is chunked\nline 2\nline3\nread 3 lines')
+        fd.close()
+
     def test_chunked_readline_wsgi_override_minimum_chunk_size(self):
 
         fd = self.connect()
